* { margin: 0; padding: 0; }

body {
	font-family: monospace;
	font-size: 1em;
	margin: 2% auto;
	max-width: 54em;
}

ul { margin-left: 1em; }
a { color: #009900; }
.Sh { font-size: 1em; padding-top: 1em; padding-bottom: 1em; }
.foot { padding-top: 1em; }

table.head, table.foot { width: 100%; }
td.head-rtitle, td.foot-os { text-align: right; }
td.head-vol { text-align: center; }
div.Pp { margin: 1ex 0ex; }
div.Nd, div.Bf, div.Op { display: inline; }
span.Pa, span.Ad { font-style: italic; }
span.Ms { font-weight: bold; }
dl.Bl-diag > dt { font-weight: bold; }
code.Nm, code.Fl, code.Cm, code.Ic, code.In, code.Fd, code.Fn,
code.Cd { font-weight: bold; font-family: inherit; }
